/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/* e-book-backend-tdb.c - tdb contact backend.
 *
 * Copyright (C) 2005 Tollef Fog Heen <tfheen@err.no>
 *
 * Copyright (C) 2005 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Tollef Fog Heen <tfheen@err.no>
 *
 * Based on work by: Nat Friedman <nat@novell.com>
 *                   Chris Toshok <toshok@ximian.com>
 *                   Hans Petter Jansson <hpj@novell.com>
 */

#ifdef HAVE_CONFIG_H
#include <config.h> 
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <tdb.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <glib/gi18n-lib.h>
#include <libedataserver/e-dbhash.h>
#include <libedataserver/e-util.h>
#include <libebook/e-contact.h>
#include <syslog.h>
#include <libedata-book/e-book-backend-sexp.h>
#include <libedata-book/e-book-backend-summary.h>
#include <libedata-book/e-data-book.h>
#include <libedata-book/e-data-book-view.h>
#include "e-book-backend-tdb.h"

#define d(x) x

#define CHANGES_DB_SUFFIX ".changes.db"

#define E_BOOK_BACKEND_TDB_VERSION_NAME "PAS-DB-VERSION"
#define E_BOOK_BACKEND_TDB_VERSION "0.2"

#define PAS_ID_PREFIX "pas-id-"
#define SUMMARY_FLUSH_TIMEOUT 5000

static EBookBackendSyncClass *e_book_backend_tdb_parent_class;

struct _EBookBackendTDBPrivate {
	char     *dirname;
	char     *filename;
	char     *summary_filename;
	TDB_CONTEXT *tdb_db;
	EBookBackendSummary *summary;
};

static void
string_to_tdb(const char *str, TDB_DATA *tdb_data)
{
	memset (tdb_data, 0, sizeof (*tdb_data));
	tdb_data->dptr = g_strdup(str);
	tdb_data->dsize = strlen (str) + 1;
}

static EContact*
create_contact (char *uid, const char *vcard)
{
	EContact *contact = e_contact_new_from_vcard (vcard);
	if (!e_contact_get_const (contact, E_CONTACT_UID))
		e_contact_set (contact, E_CONTACT_UID, uid);

	return contact;
}

static int _traverse_build_summary(TDB_CONTEXT *context, TDB_DATA key,
				   TDB_DATA value, void *data) {
	EBookBackendTDBPrivate *bfpriv = (EBookBackendTDBPrivate *) data;

	if (key.dsize != strlen(E_BOOK_BACKEND_TDB_VERSION_NAME) + 1
	    || strcmp (value.dptr, E_BOOK_BACKEND_TDB_VERSION_NAME)) {
		EContact *contact = create_contact (key.dptr, value.dptr);
		e_book_backend_summary_add_contact (bfpriv->summary, contact);
		g_object_unref (contact);
	}
	return 0;
}

static void
build_summary (EBookBackendTDBPrivate *bfpriv)
{
	TDB_CONTEXT    *db = bfpriv->tdb_db;

	tdb_traverse(db, _traverse_build_summary, bfpriv);
}

static char *
e_book_backend_tdb_create_unique_id (void)
{
	/* use a 32 counter and the 32 bit timestamp to make an id.
	   it's doubtful 2^32 id's will be created in a second, so we
	   should be okay. */
	static guint c = 0;
	return g_strdup_printf (PAS_ID_PREFIX "%08lX%08X", time(NULL), c++);
}

static void 
set_revision (EContact *contact)
{
	char time_string[25] = {0};
	const struct tm *tm = NULL;
	struct timeval tv;

	if (!gettimeofday (&tv, NULL))
		tm = gmtime (&tv.tv_sec);
	if (tm)
		strftime (time_string, 100, "%Y-%m-%dT%H:%M:%SZ", tm);
	e_contact_set (contact, E_CONTACT_REV, time_string);
	
}

static EContact *
do_create(EBookBackendTDB  *bf,
	  const char      *vcard_req)
{
	TDB_CONTEXT    *db = bf->priv->tdb_db;
	TDB_DATA       id_dbt, vcard_dbt;
	int db_error;
	char           *id;
	EContact       *contact;
	char           *vcard;
	const char *rev;
	
	id = e_book_backend_tdb_create_unique_id ();

	string_to_tdb (id, &id_dbt);

	contact = e_contact_new_from_vcard (vcard_req);
	e_contact_set (contact, E_CONTACT_UID, id);
	rev = e_contact_get_const (contact,  E_CONTACT_REV);
	if (!(rev && *rev))
		set_revision (contact);

	vcard = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);

	string_to_tdb (vcard, &vcard_dbt);

	db_error = tdb_store(db, id_dbt, vcard_dbt, TDB_REPLACE);

	g_free (vcard);

	if (0 != db_error) {
		g_warning ("db->put failed with %d", db_error);
		g_object_unref (contact);
		contact = NULL;
	}

	g_free (id);
	return contact;
}

static EBookBackendSyncStatus
e_book_backend_tdb_create_contact (EBookBackendSync *backend,
				    EDataBook *book,
				    guint32 opid,
				    const char *vcard,
				    EContact **contact)
{
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);

	*contact = do_create (bf, vcard);
	if (*contact) {
		e_book_backend_summary_add_contact (bf->priv->summary, *contact);
		return GNOME_Evolution_Addressbook_Success;
	}
	else {
		/* XXX need a different call status for this case, i
                   think */
		return GNOME_Evolution_Addressbook_ContactNotFound;
	}
}

static EBookBackendSyncStatus
e_book_backend_tdb_remove_contacts (EBookBackendSync *backend,
				     EDataBook *book,
				     guint32 opid,
				     GList *id_list,
				     GList **ids)
{
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);
	TDB_CONTEXT   *db = bf->priv->tdb_db;
	TDB_DATA       id_dbt;
	int            db_error;
	char          *id;
	GList         *l;
	GList         *removed_cards = NULL;
	GNOME_Evolution_Addressbook_CallStatus rv = GNOME_Evolution_Addressbook_Success;

	for (l = id_list; l; l = l->next) {
		id = l->data;

		string_to_tdb (id, &id_dbt);

		db_error = tdb_delete (db, id_dbt);
		if (0 != db_error) {
			rv = GNOME_Evolution_Addressbook_ContactNotFound;
			continue;
		}

		removed_cards = g_list_prepend (removed_cards, id);
	}

	*ids = removed_cards;

	for (l = removed_cards; l; l = l->next) {
		char *id = l->data;
		e_book_backend_summary_remove_contact (bf->priv->summary, id);
	}

	return rv;
}

static EBookBackendSyncStatus
e_book_backend_tdb_modify_contact (EBookBackendSync *backend,
				    EDataBook *book,
				    guint32 opid,
				    const char *vcard,
				    EContact **contact)
{
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);
	TDB_CONTEXT     *db = bf->priv->tdb_db;
	TDB_DATA        id_dbt, vcard_dbt;
	int            db_error;
	char          *id, *lookup_id;
	char          *vcard_with_rev;

	*contact = e_contact_new_from_vcard (vcard);
	id = e_contact_get(*contact, E_CONTACT_UID);

	if (id == NULL)
		return GNOME_Evolution_Addressbook_OtherError;

	lookup_id = id;

	string_to_tdb (lookup_id, &id_dbt);

	/* get the old ecard - the one that's presently in the db */
	vcard_dbt = tdb_fetch (db, id_dbt);
	if (vcard_dbt.dptr == NULL)
		return GNOME_Evolution_Addressbook_ContactNotFound;

	free (vcard_dbt.dptr);

	/* update the revisio (modified time of contact) */
	set_revision (*contact);
	vcard_with_rev = e_vcard_to_string (E_VCARD (*contact), EVC_FORMAT_VCARD_30);
	
	string_to_tdb (vcard_with_rev, &vcard_dbt);	

	db_error = tdb_store (db, id_dbt, vcard_dbt, TDB_MODIFY);

	if (0 == db_error) {

		e_book_backend_summary_remove_contact (bf->priv->summary, id);
		e_book_backend_summary_add_contact (bf->priv->summary, *contact);
	}
	g_free (id);
	g_free (vcard_with_rev);

	if (0 == db_error)
		return GNOME_Evolution_Addressbook_Success;
	else
		return GNOME_Evolution_Addressbook_ContactNotFound;
}

static EBookBackendSyncStatus
e_book_backend_tdb_get_contact (EBookBackendSync *backend,
				 EDataBook *book,
				 guint32 opid,
				 const char *id,
				 char **vcard)
{
	EBookBackendTDB *bf;
	TDB_CONTEXT     *db;
	TDB_DATA        id_dbt, vcard_dbt;

	bf = E_BOOK_BACKEND_TDB (e_data_book_get_backend (book));
	db = bf->priv->tdb_db;

	string_to_tdb (id, &id_dbt);
	memset (&vcard_dbt, 0, sizeof (vcard_dbt));

	vcard_dbt = tdb_fetch (db, id_dbt);

	if (vcard_dbt.dptr != NULL) {
		*vcard = g_strdup (vcard_dbt.dptr);
		printf ("backend %s\n", *vcard);
		free (vcard_dbt.dptr);
		return GNOME_Evolution_Addressbook_Success;
	} else {
		*vcard = g_strdup ("");
		return GNOME_Evolution_Addressbook_ContactNotFound;
	}
}

static int _traverse_append_matching_contacts(TDB_CONTEXT *context, TDB_DATA key,
						  TDB_DATA value, void *data) {
	struct traverse_append_matching_contacts_return {
		gboolean search_needed;
		EBookBackendSExp *card_sexp;
		GList *contact_list;
	};
	struct traverse_append_matching_contacts_return *r = data;
	/* don't include the version in the list of cards */
	if (key.dsize != strlen(E_BOOK_BACKEND_TDB_VERSION_NAME) + 1
	    || strcmp (key.dptr, E_BOOK_BACKEND_TDB_VERSION_NAME)) {
		
		if ((!r->search_needed) || (r->card_sexp != NULL && e_book_backend_sexp_match_vcard  (r->card_sexp, value.dptr))) {
			r->contact_list = g_list_append (r->contact_list, g_strdup (value.dptr));
		}
	}
	return 0;
}

static EBookBackendSyncStatus
e_book_backend_tdb_get_contact_list (EBookBackendSync *backend,
				      EDataBook *book,
				      guint32 opid,
				      const char *query,
				      GList **contacts)
{
	d(syslog(LOG_ERR | LOG_USER, "=> e_book_backend_tdb_get_contact_list"));
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);
	TDB_CONTEXT     *db = bf->priv->tdb_db;
	TDB_DATA  id_dbt, vcard_dbt;
	const char *search = query;
	EBookBackendSyncStatus status;
	struct traverse_append_matching_contacts_return {
		gboolean search_needed;
		EBookBackendSExp *card_sexp;
		GList *contact_list;
	} traverse_append_return_val;

	memset (&traverse_append_return_val, 0, sizeof (struct traverse_append_matching_contacts_return));
	
	d(g_warning("e_book_backend_tdb_get_contact_list (%s)\n", search));
	status = GNOME_Evolution_Addressbook_Success;
	if (e_book_backend_summary_is_summary_query (bf->priv->summary, search)) {
	
		/* do a summary query */
		GPtrArray *ids = e_book_backend_summary_search (bf->priv->summary, search);
		int i;

		for (i = 0; i < ids->len; i ++) {
			char *id = g_ptr_array_index (ids, i);
			string_to_tdb (id, &id_dbt);
			memset (&vcard_dbt, 0, sizeof (vcard_dbt));

			vcard_dbt = tdb_fetch (db, id_dbt);
			if (vcard_dbt.dptr != NULL) {
				traverse_append_return_val.contact_list = 
					g_list_append (traverse_append_return_val.contact_list, 
						       g_strdup (vcard_dbt.dptr));
				free(vcard_dbt.dptr);
			} else { 
				status = GNOME_Evolution_Addressbook_OtherError ;
				break;
			}
		}
		g_ptr_array_free (ids, TRUE);
		
	} else {
		traverse_append_return_val.search_needed = TRUE;
		if (!strcmp (search, "(contains \"x-evolution-any-field\" \"\")"))
			traverse_append_return_val.search_needed = FALSE;

		traverse_append_return_val.card_sexp = e_book_backend_sexp_new (search);
		if (!traverse_append_return_val.card_sexp) {
			/* XXX this needs to be an invalid query error of some sort*/
			return GNOME_Evolution_Addressbook_ContactNotFound;
		}

		tdb_traverse(db, _traverse_append_matching_contacts, &traverse_append_return_val);
		status = GNOME_Evolution_Addressbook_Success;

	}
	// no delete function, evidently.
	// e_book_backend_sexp_free(traverse_append_return_val.card_sexp);
	
	*contacts = traverse_append_return_val.contact_list;
	d(syslog(LOG_ERR | LOG_USER, "<= e_book_backend_tdb_get_contact_list"));

	return status;
}

typedef struct {
	EBookBackendTDB *bf;
	GCond *cond;
	GMutex *mutex;
	GThread *thread;
	gboolean stopped;
} TDBBackendSearchClosure;

static void
closure_destroy (TDBBackendSearchClosure *closure)
{
	d(printf ("destroying search closure\n"));
	g_cond_free (closure->cond);
	g_mutex_free (closure->mutex);
	g_free (closure);
}

static TDBBackendSearchClosure*
init_closure (EDataBookView *book_view, EBookBackendTDB *bf)
{
	TDBBackendSearchClosure *closure = g_new (TDBBackendSearchClosure, 1);

	closure->bf = bf;
	closure->mutex = g_mutex_new ();
	closure->cond = g_cond_new ();
	closure->thread = NULL;
	closure->stopped = FALSE;
	g_object_set_data_full (G_OBJECT (book_view), "EBookBackendTDB.BookView::closure",
				closure, (GDestroyNotify)closure_destroy);

	return closure;
}

static TDBBackendSearchClosure*
get_closure (EDataBookView *book_view)
{
	return g_object_get_data (G_OBJECT (book_view), "EBookBackendTDB.BookView::closure");
}

static gpointer
book_view_thread (gpointer data)
{
	EDataBookView *book_view = data;
	TDBBackendSearchClosure *closure = get_closure (book_view);
	EBookBackendTDB *bf = closure->bf;
	const char *query;
	TDB_CONTEXT  *db;
	TDB_DATA id_dbt, vcard_dbt;
	gboolean stopped = FALSE;

	d(printf ("starting initial population of book view\n"));

	/* ref the book view because it'll be removed and unrefed
	   when/if it's stopped */
	bonobo_object_ref (book_view);
	
	db = bf->priv->tdb_db;
	query = e_data_book_view_get_card_query (book_view);

	if ( ! strcmp (query, "(contains \"x-evolution-any-field\" \"\")"))
		e_data_book_view_notify_status_message (book_view, _("Loading..."));
	else
		e_data_book_view_notify_status_message (book_view, _("Searching..."));

	d(printf ("signalling parent thread\n"));
	g_mutex_lock (closure->mutex);
	g_cond_signal (closure->cond);
	g_mutex_unlock (closure->mutex);

	if (e_book_backend_summary_is_summary_query (bf->priv->summary, query)) {
		/* do a summary query */
		GPtrArray *ids = e_book_backend_summary_search (bf->priv->summary, e_data_book_view_get_card_query (book_view));
		int i;

		for (i = 0; i < ids->len; i ++) {
			char *id = g_ptr_array_index (ids, i);
			printf ("**** id %s\n", id);
			g_mutex_lock (closure->mutex);
			stopped = closure->stopped;
			g_mutex_unlock (closure->mutex);

			if (stopped)
				break;

			string_to_tdb (id, &id_dbt);
			memset (&vcard_dbt, 0, sizeof (vcard_dbt));

			vcard_dbt = tdb_fetch (db, id_dbt);

			if (vcard_dbt.dptr != NULL) {
				EContact *contact = create_contact (id_dbt.dptr, vcard_dbt.dptr);
				/* notify_update will check if it matches for us */
				e_data_book_view_notify_update (book_view, contact);
				
				g_object_unref (contact);

				free (vcard_dbt.dptr);
			}
			else {
				g_warning ("tdb_fetch returned %s", tdb_errorstr(db));
			}
		}

		g_ptr_array_free (ids, TRUE);
	}
	else {
		/* iterate over the db and do the query there */
		
		memset (&id_dbt, 0, sizeof (id_dbt));
		memset (&vcard_dbt, 0, sizeof (vcard_dbt));

		tdb_lockall(db);

		id_dbt = tdb_firstkey(db);
		while (id_dbt.dptr != NULL) {
			
			vcard_dbt = tdb_fetch(db, id_dbt);

			g_mutex_lock (closure->mutex);
			stopped = closure->stopped;
			g_mutex_unlock (closure->mutex);

			if (stopped)
				break;

			/* don't include the version in the list of cards */
			if (strcmp (id_dbt.dptr, E_BOOK_BACKEND_TDB_VERSION_NAME)) {
				EContact *contact = create_contact (id_dbt.dptr, vcard_dbt.dptr);
				/* notify_update will check if it matches for us */
				e_data_book_view_notify_update (book_view, contact);
				g_object_unref (contact);
			}

			id_dbt = tdb_nextkey(db, id_dbt);
		}

		tdb_unlockall(db);

	}

	if (!stopped)
		e_data_book_view_notify_complete (book_view, GNOME_Evolution_Addressbook_Success);

	/* unref the */
	bonobo_object_unref (book_view);

	d(printf ("finished population of book view\n"));

	return NULL;
}

static void
e_book_backend_tdb_start_book_view (EBookBackend  *backend,
				     EDataBookView *book_view)
{
	TDBBackendSearchClosure *closure = init_closure (book_view, E_BOOK_BACKEND_TDB (backend));

	g_mutex_lock (closure->mutex);

	d(printf ("starting book view thread\n"));
	closure->thread = g_thread_create (book_view_thread, book_view, TRUE, NULL);

	g_cond_wait (closure->cond, closure->mutex);

	/* at this point we know the book view thread is actually running */
	g_mutex_unlock (closure->mutex);
	d(printf ("returning from start_book_view\n"));
}

static void
e_book_backend_tdb_stop_book_view (EBookBackend  *backend,
				    EDataBookView *book_view)
{
	TDBBackendSearchClosure *closure = get_closure (book_view);
	gboolean need_join = FALSE;

	d(printf ("stopping query\n"));
	g_mutex_lock (closure->mutex);
	if (!closure->stopped)
		need_join = TRUE;
	closure->stopped = TRUE;
	g_mutex_unlock (closure->mutex);

	if (need_join)
		g_thread_join (closure->thread);
}

typedef struct {
	TDB_CONTEXT *db;

	GList *add_cards;
	GList *add_ids;
	GList *mod_cards;
	GList *mod_ids;
	GList *del_ids;
	GList *del_cards;
} EBookBackendTDBChangeContext;

static void
e_book_backend_tdb_changes_foreach_key (const char *key, gpointer user_data)
{
	EBookBackendTDBChangeContext *ctx = user_data;
	TDB_CONTEXT *db = ctx->db;
	TDB_DATA    id_dbt, vcard_dbt;
	
	string_to_tdb (key, &id_dbt);
	memset (&vcard_dbt, 0, sizeof (vcard_dbt));

	vcard_dbt = tdb_fetch (db, id_dbt);
	
	if (vcard_dbt.dptr != NULL) {
		EContact *contact;
		char *id = id_dbt.dptr;
		char *vcard_string;

		contact = e_contact_new ();
		e_contact_set (contact, E_CONTACT_UID, id);
		
		vcard_string = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);

		ctx->del_ids = g_list_append (ctx->del_ids,
					      g_strdup (id));
		ctx->del_cards = g_list_append (ctx->del_cards,
						vcard_string);

		g_object_unref (contact);
		
		free (vcard_dbt.dptr);
	}
}

static EBookBackendSyncStatus
e_book_backend_tdb_get_changes (EBookBackendSync *backend,
				 EDataBook *book,
				 guint32 opid,
				 const char *change_id,
				 GList **changes_out)
{
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);
	int     db_error = 0;
	TDB_DATA id_dbt, vcard_dbt;
	char    *filename;
	EDbHash *ehash;
	GList *i, *v;
	TDB_CONTEXT      *db = bf->priv->tdb_db;

	GList *changes = NULL;
	EBookBackendTDBChangeContext ctx;
	EBookBackendSyncStatus result;

	memset (&id_dbt, 0, sizeof (id_dbt));
	memset (&vcard_dbt, 0, sizeof (vcard_dbt));

	memset (&ctx, 0, sizeof (ctx));

	ctx.db = db;

	/* Find the changed ids */
	filename = g_strdup_printf ("%s/%s" CHANGES_DB_SUFFIX, bf->priv->dirname, change_id);
	ehash = e_dbhash_new (filename);
	g_free (filename);

	db_error = tdb_lockall(db);
	
	if (db_error != 0) {
		g_warning ("e_book_backend_tdb_changes: error building list\n");
	} else {
		id_dbt = tdb_firstkey(db);

		while (id_dbt.dptr != NULL) {
			
			vcard_dbt = tdb_fetch(db, id_dbt);

			/* don't include the version in the list of cards */
			if (id_dbt.dsize != strlen(E_BOOK_BACKEND_TDB_VERSION_NAME) + 1
			    || strcmp (id_dbt.dptr, E_BOOK_BACKEND_TDB_VERSION_NAME)) {
				EContact *contact;
				char *id = id_dbt.dptr;
				char *vcard_string;
				
				/* Remove fields the user can't change
				 * and can change without the rest of the
				 * card changing 
				 */
				contact = create_contact (id_dbt.dptr, vcard_dbt.dptr);

#if notyet
				g_object_set (card, "last_use", NULL, "use_score", 0.0, NULL);
#endif
				vcard_string = e_vcard_to_string (E_VCARD (contact), EVC_FORMAT_VCARD_30);
				g_object_unref (contact);
				
				/* check what type of change has occurred, if any */
				switch (e_dbhash_compare (ehash, id, vcard_string)) {
				case E_DBHASH_STATUS_SAME:
					break;
				case E_DBHASH_STATUS_NOT_FOUND:
					ctx.add_cards = g_list_append (ctx.add_cards, vcard_string);
					ctx.add_ids = g_list_append (ctx.add_ids, g_strdup(id));
					break;
				case E_DBHASH_STATUS_DIFFERENT:
					ctx.mod_cards = g_list_append (ctx.mod_cards, vcard_string);
					ctx.mod_ids = g_list_append (ctx.mod_ids, g_strdup(id));
					break;
				}
			}
			free(vcard_dbt.dptr);
			free(id_dbt.dptr);
			id_dbt = tdb_nextkey(db, id_dbt);
		}
		free(id_dbt.dptr);
		
		tdb_unlockall(db);
	}

   	e_dbhash_foreach_key (ehash, (EDbHashFunc)e_book_backend_tdb_changes_foreach_key, &ctx);

	/* Send the changes */
	/* Update the hash and build our changes list */
	for (i = ctx.add_ids, v = ctx.add_cards; i != NULL; i = i->next, v = v->next){
		char *id = i->data;
		char *vcard = v->data;

		e_dbhash_add (ehash, id, vcard);
		changes = g_list_prepend (changes,
					  e_book_backend_change_add_new (vcard));
		
		g_free (i->data);
		g_free (v->data);
	}	
	for (i = ctx.mod_ids, v = ctx.mod_cards; i != NULL; i = i->next, v = v->next){
		char *id = i->data;
		char *vcard = v->data;
		
		e_dbhash_add (ehash, id, vcard);
		changes = g_list_prepend (changes,
					  e_book_backend_change_modify_new (vcard));
		
		g_free (i->data);
		g_free (v->data);		
	}
	for (i = ctx.del_ids, v = ctx.del_cards; i != NULL; i = i->next, v = v->next){	
		char *id = i->data;
		char *vcard = v->data;
		
		e_dbhash_remove (ehash, id);
		
		changes = g_list_prepend (changes,
					  e_book_backend_change_delete_new (vcard));
		g_free (i->data);
		g_free (v->data);
	}
	
	e_dbhash_write (ehash);
	
	result = GNOME_Evolution_Addressbook_Success;
	*changes_out = changes;

	e_dbhash_destroy (ehash);
	
	return GNOME_Evolution_Addressbook_Success;
}

static char *
e_book_backend_tdb_extract_path_from_uri (const char *uri)
{
	g_assert (strncasecmp (uri, "tdb://", 6) == 0);

	return g_strdup (uri + 6);
}

static EBookBackendSyncStatus
e_book_backend_tdb_authenticate_user (EBookBackendSync *backend,
				       EDataBook *book,
				       guint32 opid,
				       const char *user,
				       const char *passwd,
				       const char *auth_method)
{
	return GNOME_Evolution_Addressbook_Success;
}

static EBookBackendSyncStatus
e_book_backend_tdb_get_required_fields (EBookBackendSync *backend,
					  EDataBook *book,
					  guint32 opid,
					  GList **fields_out)
{
	GList *fields = NULL;
	
	fields = g_list_append (fields , g_strdup(e_contact_field_name (E_CONTACT_FILE_AS)));
	*fields_out = fields;
	return GNOME_Evolution_Addressbook_Success;
}


static EBookBackendSyncStatus
e_book_backend_tdb_get_supported_fields (EBookBackendSync *backend,
					  EDataBook *book,
					  guint32 opid,
					  GList **fields_out)
{
	GList *fields = NULL;
	int i;

	/* XXX we need a way to say "we support everything", since the
	   file backend does */
	for (i = 1; i < E_CONTACT_FIELD_LAST; i ++)
		fields = g_list_append (fields, g_strdup (e_contact_field_name (i)));

	*fields_out = fields;
	return GNOME_Evolution_Addressbook_Success;
}

#ifdef CREATE_DEFAULT_VCARD
# include <libedata-book/ximian-vcard.h>
#endif

static GNOME_Evolution_Addressbook_CallStatus
e_book_backend_tdb_load_source (EBookBackend           *backend,
				 ESource                *source,
				 gboolean                only_if_exists)
{
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);
	char           *dirname, *filename;
	gboolean        writable = FALSE;
	TDB_CONTEXT *db;
	time_t db_mtime;
	struct stat sb;
	gchar *uri;

	uri = e_source_get_uri (source);

	dirname = e_book_backend_tdb_extract_path_from_uri (uri);
	filename = g_build_filename (dirname, "addressbook.tdb", NULL);
	g_free (uri);

	/* FIXME; Do we care that we have a race condition here? */
	if (!g_file_test (filename, G_FILE_TEST_EXISTS)) { /* New database */
		int rv;

		rv = e_util_mkdir_hier (dirname, 0777);
		if (rv == -1 && errno != EEXIST) {
			g_warning ("failed to make directory %s: %s", dirname, strerror (errno));
			if (errno == EACCES || errno == EPERM)
				return GNOME_Evolution_Addressbook_PermissionDenied;
			else
				return GNOME_Evolution_Addressbook_OtherError;
		}

		db = tdb_open(filename, 0, 0, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
		if (db == NULL) {
			g_warning ("tdb_open failed");
			return GNOME_Evolution_Addressbook_OtherError;
		}
#ifdef CREATE_DEFAULT_VCARD
		EContact *contact;
		
		contact = do_create(bf, XIMIAN_VCARD);
		/* XXX check errors here */
		g_object_unref (contact);
#endif
		writable = TRUE;
	} else {
		if (access( filename, R_OK | W_OK) == 0) {
			db = tdb_open(filename, 0, 0, O_RDWR, S_IRUSR | S_IWUSR);
			if (db == NULL) {
				g_warning ("tdb_open failed");
				return GNOME_Evolution_Addressbook_OtherError;
			}
			writable = TRUE;
		} else if (access( filename, R_OK) == 0) {
			db = tdb_open(filename, 0, 0, O_RDONLY, S_IRUSR | S_IWUSR);
			if (db == NULL) {
				g_warning ("tdb_open on read-only addressbook failed");
				return GNOME_Evolution_Addressbook_OtherError;
			}
		} else {
			g_warning ("addressbook not readable");
			return GNOME_Evolution_Addressbook_OtherError;
		}
	}
	bf->priv->tdb_db = db;

	g_free (bf->priv->dirname);
	g_free (bf->priv->filename);
	bf->priv->dirname = dirname;
	bf->priv->filename = filename;

	if (stat (bf->priv->filename, &sb) == -1) {
		tdb_close (db);
		bf->priv->tdb_db = NULL;
		g_warning ("stat(%s) failed", bf->priv->filename);
		return GNOME_Evolution_Addressbook_OtherError;
	}
	db_mtime = sb.st_mtime;

	g_free (bf->priv->summary_filename);
	bf->priv->summary_filename = g_strconcat (bf->priv->filename, ".summary", NULL);
	bf->priv->summary = e_book_backend_summary_new (bf->priv->summary_filename, SUMMARY_FLUSH_TIMEOUT);

	if (e_book_backend_summary_is_up_to_date (bf->priv->summary, db_mtime) == FALSE
	    || e_book_backend_summary_load (bf->priv->summary) == FALSE ) {
		build_summary (bf->priv);
	}

	e_book_backend_set_is_loaded (backend, TRUE);
	e_book_backend_set_is_writable (backend, writable);
	return GNOME_Evolution_Addressbook_Success;
}

static gboolean
select_changes (const char *name)
{
	char *p;

	if (strlen (name) < strlen (CHANGES_DB_SUFFIX))
		return FALSE;

	p = strstr (name, CHANGES_DB_SUFFIX);
	if (!p)
		return FALSE;

	if (strlen (p) != strlen (CHANGES_DB_SUFFIX))
		return FALSE;

	return TRUE;
}

static EBookBackendSyncStatus
e_book_backend_tdb_remove (EBookBackendSync *backend,
			    EDataBook        *book,
			    guint32           opid)
{
	EBookBackendTDB *bf = E_BOOK_BACKEND_TDB (backend);
	GDir *dir;

	if (-1 == unlink (bf->priv->filename)) {
		if (errno == EACCES || errno == EPERM)
			return GNOME_Evolution_Addressbook_PermissionDenied;
		else
			return GNOME_Evolution_Addressbook_OtherError;
	}

	/* unref the summary before we remove the file so it's not written out again */
	g_object_unref (bf->priv->summary);
	bf->priv->summary = NULL;
	if (-1 == unlink (bf->priv->summary_filename))
		g_warning ("failed to remove summary file `%s`: %s", bf->priv->summary_filename, strerror (errno));

	dir = g_dir_open (bf->priv->dirname, 0, NULL);
	if (dir) {
		const char *name;

		while ((name = g_dir_read_name (dir))) {
			if (select_changes (name)) {
				char *full_path = g_build_filename (bf->priv->dirname, name, NULL);
				if (-1 == unlink (full_path)) {
					g_warning ("failed to remove change db `%s': %s", full_path, strerror (errno));
				}
				g_free (full_path);
			}
		}

		g_dir_close (dir);
	}

	if (-1 == rmdir (bf->priv->dirname))
		g_warning ("failed to remove directory `%s`: %s", bf->priv->dirname, strerror (errno));

	/* we may not have actually succeeded in removing the
	   backend's files/dirs, but there's nothing we can do about
	   it here..  the only time we should return failure is if we
	   failed to remove the actual data.  a failure should mean
	   that the addressbook is still valid */
	return GNOME_Evolution_Addressbook_Success;
}

static char *
e_book_backend_tdb_get_static_capabilities (EBookBackend *backend)
{
	return g_strdup("local,do-initial-query,bulk-removes,contact-lists");
}

static GNOME_Evolution_Addressbook_CallStatus
e_book_backend_tdb_cancel_operation (EBookBackend *backend, EDataBook *book)
{
	return GNOME_Evolution_Addressbook_CouldNotCancel;
}
static void 
e_book_backend_tdb_set_mode (EBookBackend *backend,  GNOME_Evolution_Addressbook_BookMode mode)
{
	if (e_book_backend_is_loaded (backend)) {
		e_book_backend_notify_writable (backend, TRUE);
		e_book_backend_notify_connection_status (backend, TRUE);
	}
}
static gboolean
e_book_backend_tdb_construct (EBookBackendTDB *backend)
{
	g_assert (backend != NULL);
	g_assert (E_IS_BOOK_BACKEND_TDB (backend));

	if (! e_book_backend_construct (E_BOOK_BACKEND (backend)))
		return FALSE;

	return TRUE;
}

/**
 * e_book_backend_tdb_new:
 */
EBookBackend *
e_book_backend_tdb_new (void)
{
	EBookBackendTDB *backend;

	backend = g_object_new (E_TYPE_BOOK_BACKEND_TDB, NULL);

	if (! e_book_backend_tdb_construct (backend)) {
		g_object_unref (backend);

		return NULL;
	}

	return E_BOOK_BACKEND (backend);
}

static void
e_book_backend_tdb_dispose (GObject *object)
{
	EBookBackendTDB *bf;

	bf = E_BOOK_BACKEND_TDB (object);

	if (bf->priv) {
		if (bf->priv->tdb_db)
			tdb_close (bf->priv->tdb_db);

		if (bf->priv->summary)
			g_object_unref(bf->priv->summary);
		g_free (bf->priv->filename);
		g_free (bf->priv->dirname);
		g_free (bf->priv->summary_filename);

		g_free (bf->priv);
		bf->priv = NULL;
	}

	G_OBJECT_CLASS (e_book_backend_tdb_parent_class)->dispose (object);	
}

static void
e_book_backend_tdb_class_init (EBookBackendTDBClass *klass)
{
	GObjectClass    *object_class = G_OBJECT_CLASS (klass);
	EBookBackendSyncClass *sync_class;
	EBookBackendClass *backend_class;

	e_book_backend_tdb_parent_class = g_type_class_peek_parent (klass);

	sync_class = E_BOOK_BACKEND_SYNC_CLASS (klass);
	backend_class = E_BOOK_BACKEND_CLASS (klass);

	/* Set the virtual methods. */
	backend_class->load_source             = e_book_backend_tdb_load_source;
	backend_class->get_static_capabilities = e_book_backend_tdb_get_static_capabilities;
	backend_class->start_book_view         = e_book_backend_tdb_start_book_view;
	backend_class->stop_book_view          = e_book_backend_tdb_stop_book_view;
	backend_class->cancel_operation        = e_book_backend_tdb_cancel_operation;
	backend_class->set_mode                = e_book_backend_tdb_set_mode;
	sync_class->remove_sync                = e_book_backend_tdb_remove;
	sync_class->create_contact_sync        = e_book_backend_tdb_create_contact;
	sync_class->remove_contacts_sync       = e_book_backend_tdb_remove_contacts;
	sync_class->modify_contact_sync        = e_book_backend_tdb_modify_contact;
	sync_class->get_contact_sync           = e_book_backend_tdb_get_contact;
	sync_class->get_contact_list_sync      = e_book_backend_tdb_get_contact_list;
	sync_class->get_changes_sync           = e_book_backend_tdb_get_changes;
	sync_class->authenticate_user_sync     = e_book_backend_tdb_authenticate_user;
	sync_class->get_supported_fields_sync  = e_book_backend_tdb_get_supported_fields;
	sync_class->get_required_fields_sync   = e_book_backend_tdb_get_required_fields;
	

	object_class->dispose = e_book_backend_tdb_dispose;
}

static void
e_book_backend_tdb_init (EBookBackendTDB *backend)
{
	EBookBackendTDBPrivate *priv;

	priv             = g_new0 (EBookBackendTDBPrivate, 1);

	backend->priv = priv;
}

/**
 * e_book_backend_tdb_get_type:
 */
GType
e_book_backend_tdb_get_type (void)
{
	static GType type = 0;

	if (! type) {
		GTypeInfo info = {
			sizeof (EBookBackendTDBClass),
			NULL, /* base_class_init */
			NULL, /* base_class_finalize */
			(GClassInitFunc)  e_book_backend_tdb_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (EBookBackendTDB),
			0,    /* n_preallocs */
			(GInstanceInitFunc) e_book_backend_tdb_init
		};

		type = g_type_register_static (E_TYPE_BOOK_BACKEND_SYNC, "EBookBackendTDB", &info, 0);
	}

	return type;
}
